#define _XOPEN_SOURCE 600

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <string.h>
#include <sys/fcntl.h>
#include <sys/select.h>
#include <errno.h>
#include <unistd.h>

#define MASTER_RANK     0

static int _log_fd = 1;

static int * _watch_fds = NULL;
static int _watch_fds_count = 0;

void logprintf(char * fmt, ...)
{
 va_list ap;
 time_t t; 
 char buf[1024];
 
 if (_log_fd < 0)
    return;
 
 t = time(NULL);
 strftime(buf, sizeof(buf), "[ %d-%m-%Y %T ] ", localtime(&t));
 if (write(_log_fd, buf, strlen(buf)) < 0)
 {
  perror("write()");
  return;
 }
 
 va_start(ap, fmt); 
 vsnprintf(buf, sizeof(buf), fmt, ap);
 va_end(ap);
 
 if (write(_log_fd, buf, strlen(buf)) < 0)
 {
  perror("write()");
  return;
 }
}

void sendlog(char * fmt, ...)
{
 va_list ap; 
 char buf[1024];  
 int id;
 
 MPI_Comm_rank(MPI_COMM_WORLD, &id);
 sprintf(buf, "node%02d: ", id);
 
 va_start(ap, fmt); 
 vsnprintf(&buf[strlen(buf)], sizeof(buf) - strlen(buf), fmt, ap);
 va_end(ap);
 
 MPI_Send(buf, strlen(buf) + 1, MPI_CHAR, MASTER_RANK, 0, MPI_COMM_WORLD);
}

void do_master(void)
{
	int hosts;
	char buf[256];
	MPI_Status status;
	
	MPI_Comm_size(MPI_COMM_WORLD, &hosts);
	logprintf("master: startuje, ilosc wezlow: %d\n", hosts - 1);
	
	/* Czekamy na wiadomosci */	
	while(1)
	{
		MPI_Recv(buf, sizeof(buf), MPI_CHAR, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
		logprintf(buf);
	}
}

void node_addfile(char * path)
{
	_watch_fds_count++;
	_watch_fds = realloc(_watch_fds, sizeof(int) * _watch_fds_count);
	_watch_fds[_watch_fds_count - 1] = open(path, O_RDONLY);
	if (_watch_fds[_watch_fds_count - 1] < 0)
	{
		_watch_fds_count--;
		sendlog("%s: %s\n", path, strerror(errno));
		_watch_fds = realloc(_watch_fds, sizeof(int) * _watch_fds_count);
	}
}

void do_node(void)
{
	int i, err;
	char buf[1024];
	fd_set fds;
	
	sendlog("startuje\n");
	
	while(1)
	{
		FD_ZERO(&fds);
		for(i=0;i<_watch_fds_count;i++)
			FD_SET(_watch_fds[i], &fds);
		err = select(FD_SETSIZE, &fds, NULL, NULL, NULL);
		if (err < 0)
		{
			sendlog("select(): %s\n", strerror(errno));
			break;
		}
		
		for(i=0;i<_watch_fds_count;i++)
		{
			if (FD_ISSET(_watch_fds[i], &fds))
			{
				memset(buf, 0, sizeof(buf));
				if (read(_watch_fds[i], buf, sizeof(buf) - 1) > 0)
					sendlog(buf);
			}
		}
	}
}

int main(int argc, char * argv[])
{
	int id;
	FILE * fp;
	char line[1024];
	
	MPI_Init(&argc,&argv);
	
	MPI_Comm_rank(MPI_COMM_WORLD, &id);
	
	/* Czytamy plik konfiguracyjny */
	if ((fp = fopen("config.txt", "r")) == NULL)
	{
		printf("Nie mozna otworzyc pliku konfiguracyjnego!\n");
		return 1;
	}
 
	while(fgets(line, sizeof(line) - 1, fp))
	{
		if ((line[0] == '#') || (line[0] == '\n')) /* Komentarz lub pusta linia */
			continue;
		line[strlen(line) - 1] = '\0'; /* Wywal \n z konca */
		if ((!strncmp("Output ", line, strlen("Output "))) && (id == MASTER_RANK))
		{
			_log_fd = open(&line[strlen("Output ")], O_WRONLY | O_APPEND);
			if (_log_fd < 0)
			{
				perror("Nie mozna otworzyc pliku wynikowego");
				return 2;
			}
		}
		else if ((!strncmp("Source ", line, strlen("Source "))) && (id != MASTER_RANK))
		{
			node_addfile(&line[strlen("Source ")]);
		}
	}
	
	
	if (id == MASTER_RANK)
		do_master();
	else
		do_node();
	
	MPI_Finalize(); 
	return 0;
}
